package com.aurellem.capture.examples;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.ByteBuffer;

import javax.sound.sampled.AudioFormat;

import org.tritonus.share.sampled.FloatSampleTools;

import com.aurellem.capture.Capture;
import com.aurellem.capture.IsoTimer;
import com.aurellem.capture.audio.CompositeSoundProcessor;
import com.aurellem.capture.audio.MultiListener;
import com.aurellem.capture.audio.SoundProcessor;
import com.aurellem.capture.audio.WaveFileWriter;
import com.jme3.app.SimpleApplication;
import com.jme3.audio.AudioNode;
import com.jme3.audio.AudioParam;
import com.jme3.audio.Listener;
import com.jme3.cinematic.MotionPath;
import com.jme3.cinematic.events.MotionTrack;
import com.jme3.input.controls.ActionListener;
import com.jme3.input.controls.MouseButtonTrigger;
import com.jme3.light.DirectionalLight;
import com.jme3.material.Material;
import com.jme3.math.ColorRGBA;
import com.jme3.math.FastMath;
import com.jme3.math.Quaternion;
import com.jme3.math.Vector3f;
import com.jme3.scene.Geometry;
import com.jme3.scene.Node;
import com.jme3.scene.Spatial;
import com.jme3.scene.shape.Box;
import com.jme3.scene.shape.Sphere;
import com.jme3.system.AppSettings;


/**
 * 
 * Demonstrates advanced use of the audio capture and recording features.
 * Multiple perspectives of the same scene are simultaneously rendered to 
 * different sound files.  
 * 
 * A key limitation of the way multiple listeners are implemented is that 
 * only 3D positioning effects are realized for listeners other than the
 * main LWJGL listener.  This means that audio effects such as environment
 * settings will *not* be heard on any auxiliary listeners, though sound 
 * attenuation will work correctly.  
 * 
 * Multiple listeners as realized here might be used to make AI entities 
 * that can each hear the world from their own perspective.  
 * 
 * @author Robert McIntyre
 *
 */

public class AdvancedAudio extends SimpleApplication {

	  public static void main(String[] args) {
		  
			AdvancedAudio app = new AdvancedAudio();
			AppSettings settings = new AppSettings(true);
			settings.setAudioRenderer("Send");
			app.setSettings(settings);
			app.setShowSettings(false);
			app.setPauseOnLostFocus(false);
			org.lwjgl.input.Mouse.setGrabbed(false);
			try {Capture.captureVideo(app, new File("/home/r/tmp/out.avi"));}
			catch (IOException e) {e.printStackTrace();}
			app.start();
	  }
	 
	 private MotionTrack motionControl;
	
	 
	 private Spatial makeEar(Node root, Vector3f position){
		 Material mat = new Material(assetManager, "Common/MatDefs/Misc/Unshaded.j3md");
		 Geometry ear = new Geometry("ear", new Box(1.0f, 1.0f, 1.0f));
		 ear.setLocalTranslation(position);
		 mat.setColor("Color", ColorRGBA.Green);
		 ear.setMaterial(mat);
		 root.attachChild(ear);
		 return ear;
	 } 
	 
	 private Geometry bell;
	 
	 private Spatial ear1;
	 //private Spatial ear2;
	 //private Spatial ear3;
	 //private Spatial ear4;
	 
	 
	 private Vector3f[] path = new Vector3f[]{
			 // loop 1
			 new Vector3f(0, 0, 0),
			 new Vector3f(0, 0, -10),
			 new Vector3f(-2, 0, -14),
			 new Vector3f(-6, 0, -20),
			 new Vector3f(0, 0, -26),
			 new Vector3f(6, 0, -20),
			 new Vector3f(0, 0, -14),
			 new Vector3f(-6, 0, -20),
			 new Vector3f(0, 0, -26),
			 new Vector3f(6, 0, -20),
			 // loop 2
			 new Vector3f(5, 0, -5),
			 new Vector3f(7, 0, 1.5f),
			 new Vector3f(14, 0, 2),
			 new Vector3f(20, 0, 6),
			 new Vector3f(26, 0, 0),
			 new Vector3f(20, 0, -6),
			 new Vector3f(14, 0, 0),
			 new Vector3f(20, 0, 6),
			 new Vector3f(26, 0, 0),
			 new Vector3f(20, 0, -6),
			 new Vector3f(14, 0, 0),
			 // loop 3
			 new Vector3f(8, 0, 7.5f),
			 new Vector3f(7, 0, 10.5f),
			 new Vector3f(6, 0, 20),
			 new Vector3f(0, 0, 26),
			 new Vector3f(-6, 0, 20),
			 new Vector3f(0, 0, 14),
			 new Vector3f(6, 0, 20),
			 new Vector3f(0, 0, 26),
			 new Vector3f(-6, 0, 20),
			 new Vector3f(0, 0, 14),
			 // begin ellipse
			 new Vector3f(16, 5, 20),
			 new Vector3f(0, 0, 26),
			 new Vector3f(-16, -10, 20),
			 new Vector3f(0, 0, 14),
			 new Vector3f(16, 20, 20),
			 new Vector3f(0, 0, 26),
			 new Vector3f(-10, -25, 10),
			 new Vector3f(-10, 0, 0),
			 // come at me bro!
			 new Vector3f(-28.00242f, 48.005623f, -34.648228f),
			 new Vector3f(0, 0 , -20),
	 };
	 
	 
	 
	 private void createScene() {
		 Material mat = new Material(assetManager, "Common/MatDefs/Misc/Unshaded.j3md");
		 bell = new Geometry( "sound-emitter" , new Sphere(15,15,1));
		 mat.setColor("Color", ColorRGBA.Blue);
		 bell.setMaterial(mat);
		 rootNode.attachChild(bell);
		 
		 DirectionalLight light = new DirectionalLight();
		 light.setDirection(new Vector3f(0, -1, 0).normalizeLocal());
		 light.setColor(ColorRGBA.White.mult(1.5f));
		 rootNode.addLight(light);

		 ear1 = makeEar(rootNode, new Vector3f(0, 0 ,20));
		 //ear2 = makeEar(rootNode, new Vector3f(0, 0 ,-20));
		 //ear3 = makeEar(rootNode, new Vector3f(20, 0 ,0));
		 //ear4 = makeEar(rootNode, new Vector3f(-20, 0 ,0));

		 MotionPath track = new MotionPath();
		 
		for (Vector3f v : path){
			track.addWayPoint(v);
		}
	
		  
		 track.setCurveTension(0.80f);

		 
		 motionControl = new MotionTrack(bell,track);
		 motionControl.setTimer(new IsoTimer(60));
	     motionControl.setDirectionType(MotionTrack.Direction.PathAndRotation);
	     motionControl.setRotation(new Quaternion().fromAngleNormalAxis(-FastMath.HALF_PI, Vector3f.UNIT_Y));
	     motionControl.setInitialDuration(20f);
	     motionControl.setSpeed(1f);
		 
		 
		 track.enableDebugShape(assetManager, rootNode);


		 positionCamera();
		
	        
	    }

	 
	 private void positionCamera(){
		this.cam.setLocation(new Vector3f(-28.00242f, 48.005623f, -34.648228f));
		// cam.setLocation(new Vector3f(0,0,-20));
		this.cam.setRotation(new Quaternion(0.3359635f, 0.34280345f, -0.13281013f, 0.8671653f));
	 }

  private AudioNode music;
  
  

   
  
  private void initAudio() {

	music = new AudioNode(assetManager, "Sound/Environment/pure.wav", false);
	
    rootNode.attachChild(music);
    audioRenderer.playSource(music);
    music.setPositional(true);
    music.setVolume(1f);
    music.setReverbEnabled(false);
    music.setMaxDistance(200.0f);
    music.setRefDistance(1f);
    music.setRolloffFactor(5f);
    audioRenderer.pauseSource(music); 
    
  }

  
  
  
  private Listener auxListener = new Listener(); 
  //public File data1 = new File("/home/r/tmp/data1.wav");
  //public File data2 = new File("/home/r/tmp/data2.wav");
  //public File data3 = new File("/home/r/tmp/data3.wav");
  //public File data4 = new File("/home/r/tmp/data4.wav");
  //public File data5 = new File("/home/r/tmp/data5.wav");
  //public File data6 = new File("/home/r/tmp/data6.wav");


  public class Dancer implements SoundProcessor {

	  Spatial entity;

	  float scale = 2;
	  String debug;
	  public Dancer(Spatial entity, String debug){
		  this.entity = entity;
		  this.debug = debug;
	  }

	  /**
	   * this method is irrelevant since there is no state to cleanup.
	   */
	  public void cleanup() {}


	  /**
	   * Dance to the beat!  This is the brain of an AI entity that 
	   * hears it's surroundings and reacts to them.
	   */
	  public void process(ByteBuffer audioSamples, int numSamples, AudioFormat format) {
		  audioSamples.clear();
		  byte[] data = new byte[numSamples];
		  float[] out = new float[numSamples];
		  audioSamples.get(data);
		  FloatSampleTools.byte2floatInterleaved(data, 0, out, 0, 
				  numSamples/format.getFrameSize(), format);
		  
		  float max = Float.NEGATIVE_INFINITY;
		  for (float f : out){if (f > max) max = f;}
		  audioSamples.clear();
		  System.out.println(debug);
		  System.out.println(max);
		  
		  
		  
		  //entity.scale(this.scale);
		  //if (this.scale == 2f){this.scale = 0.5f;}
		  //else {this.scale = 2;}
	  }


  }

  
  
  
  public void simpleInitApp() {
	this.setTimer(new IsoTimer(60));
    initAudio();
    initKeys();
	createScene();
	if (this.audioRenderer instanceof MultiListener){
		MultiListener rf = (MultiListener)this.audioRenderer;
		
		
		listener.setLocation(ear1.getLocalTranslation());
		auxListener = new Listener(listener);
		
		rf.addListener(auxListener);
		WaveFileWriter aux = null;
		WaveFileWriter main = null;

		
		try {aux = new WaveFileWriter(new File("/home/r/tmp/aux.wav"));} 
		catch (FileNotFoundException e) {e.printStackTrace();}
		
		try {main = new WaveFileWriter(new File("/home/r/tmp/main.wav"));} 
		catch (FileNotFoundException e) {e.printStackTrace();}
				
		rf.registerSoundProcessor(auxListener, 
				new CompositeSoundProcessor(new Dancer(ear1, "aux"), aux));
		
		rf.registerSoundProcessor(
				new CompositeSoundProcessor(new Dancer(ear1, "--------\nmain"), main));
	}   

    motionControl.play();
  }

 


  
  private void initKeys() {
	  inputManager.addMapping("Shoot", new MouseButtonTrigger(0));
	  inputManager.addListener(actionListener, "Shoot");
  }

  /** Defining the "Shoot" action: Play a gun sound. */
  private ActionListener actionListener = new ActionListener() {
	  @Override
	  public void onAction(String name, boolean keyPressed, float tpf) {
		  if (name.equals("Shoot") && !keyPressed) {
			  System.out.println("I'm playing! <3");
			  System.out.println(bell.getLocalTranslation().subtract(cam.getLocation()).length());	
			  bell.getMaterial().setColor("Color", ColorRGBA.randomColor());
			  audioRenderer.playSource(music);
			  System.out.println(music.getRefDistance());
			 
		  }
	  }
  };

  /** Move the listener with the camera - for 3D audio. */
  
  
  private Vector3f prevBellPos = Vector3f.ZERO;
  
  
  public void simpleUpdate(float tpf) {
	//Vector3f loc = cam.getLocation();
	//Quaternion rot = cam.getRotation();
    //listener.setLocation(loc);
    //listener.setRotation(rot);
    
    
    //listener.setLocation(cam.getLocation());
    //listener.setRotation(cam.getRotation());
    //auxListener.setLocation(loc);
    //auxListener.setRotation(rot);
    if (music.getStatus() != AudioNode.Status.Playing){
    	audioRenderer.playSource(music);
    	bell.getMaterial().setColor("Color", ColorRGBA.randomColor());
    }
    //audioRenderer.updateSourceParam(music, AudioParam.Direction);
    
    Vector3f bellVelocity = bell.getLocalTranslation().subtract(prevBellPos).mult(1.0f/tpf);
    prevBellPos = bell.getLocalTranslation();
    
    music.setLocalTranslation(bell.getLocalTranslation());
    music.setVelocity(bellVelocity);
    
    //audioRenderer.updateSourceParam(music, AudioParam.Position);
    //audioRenderer.updateSourceParam(music, AudioParam.Velocity);
    
    
    //System.out.println("main:" + listener.getVolume());
    //System.out.println("aux:" + auxListener.getVolume());
    //org.lwjgl.openal.AL10.alSourcef(1, org.lwjgl.openal.AL10.AL_MIN_GAIN, 0f);
    //org.lwjgl.openal.AL10.alSourcef(1, org.lwjgl.openal.AL10.AL_ROLLOFF_FACTOR, 5f);
    	
  }

}
